/**
* Programa Tabuleiro 2:
*	Jogo desenvolvido para o ensino de desenvolvimento de
*   jogos de computadores, utilizando-se de matriz de
*   elementos
*
*	Autor
*		Saulo Popov Zambiasi <saulopz@gmail.com>
*
*	Ultima alteracao
*		08 de Outubro de 2008.
*/

#include <stdio.h>
#include <allegro.h>

#define MAX_X 800
#define MAX_Y 600
#define VMODE GFX_AUTODETECT_WINDOWED

// Direcao
#define PARADO		0
#define DIREITA		1
#define CIMA		2
#define ESQUERDA	3
#define BAIXO		4

// Tipos de elementos
#define PAREDE		1
#define MONSTRO		2
#define AVATAR		3

/**
*   Classe Elemento - Representa cada objeto ativo ou inativo no jogo
*/

class Elemento {
public:
    int tipo;   // Tipo do elemento
    int direcao;    // Direcao da qual o elemento esta andando
    int jogada; // Cada elemento roda apenas uma vez a cada jogada do ambiente
    bool andando;   // Verifica se o elemento esta andando
    BITMAP *imagem[2];  // Imagens que representam o elemento
    // Construtor
    Elemento(int t, BITMAP* img0, BITMAP* img1) {
        tipo = t;
        jogada = 0;
        imagem[0] = img0;
        imagem[1] = img1;
        direcao = PARADO;
        andando = false;
    }
};

Elemento *mapa[16][12]; // Mapa de elementos no jogo
BITMAP *buffer; // Folha para desenhar o jogo
BITMAP *chao;   // Fundo do jogo
int passo;  // Execuo no passo 0. O Passo 1 representa a suavizao do movimento
int jogada; // Jogada atual. Para os elementos executarem apenas uma vez a cada jogada
bool sair;   // Sair do jogo?

/**
* Funcao para controle da temporizacao - allegro
*/
int tempo;
void aumenta_tempo (void) {
    tempo++;
}
END_OF_FUNCTION(aumenta_tempo);

/**
* Move o elemento para a posicao desejada
*/
bool mover_elemento(int x, int y) {
    if (mapa[x][y] == NULL) {
        return false;
    }
    int novo_x = x;
    int novo_y = y;
    switch (mapa[x][y]->direcao) {
        case CIMA: novo_y--; break;
        case BAIXO: novo_y++; break;
        case ESQUERDA: novo_x--; break;
        case DIREITA: novo_x++; break;
    }
    if ((novo_x < 0) || (novo_x >= 16) ||
        (novo_y < 0) || (novo_y >= 12)) {
        return false;
    }
    if (mapa[novo_x][novo_y] != NULL) {
        return false;
    }
    mapa[x][y]->andando = true;
    mapa[novo_x][novo_y] = mapa[x][y];
    mapa[x][y] = NULL;
    return true;
}

/**
* Executa o elemento verificando qual sua acao
*/
bool executar_elemento(int x, int y) {
    if (mapa[x][y] == NULL) {
        return false;
    }
    if (passo == 1) {
        mapa[x][y]->andando = false;
        return false;
    }
    if (mapa[x][y]->jogada == jogada) {
        return false;
    }
    mapa[x][y]->jogada = jogada;
    int direcao = PARADO;
    switch (mapa[x][y]->tipo) {
        case AVATAR:
            if (key[KEY_UP]) direcao = CIMA;
            else if (key[KEY_DOWN]) direcao = BAIXO;
            else if (key[KEY_LEFT]) direcao = ESQUERDA;
            else if (key[KEY_RIGHT]) direcao = DIREITA;
            break;
        case MONSTRO:
            direcao = (rand() % 5);
            break;
        default:
            return false;
    }
    mapa[x][y]->direcao = direcao;
    mover_elemento(x, y);
    return true;
}

/**
* Retorna a rotacao o elemento, conforme sua direcao
*/
int rotacionar_objeto(int direcao) {
    switch (direcao) {
        case DIREITA: return -32;
        case CIMA: return 128;
        case ESQUERDA: return 32;
        default: return 0;
    }
}

/**
* Desenha todos os elementos do jogo
*/
void desenhar() {
    int gira = 0;
    blit(chao, buffer, 0, 0, 0, 0, MAX_X, MAX_Y);
    for (int x = 0; x < 16; x++) {
        for (int y = 0; y < 12; y++) {
            if (mapa[x][y] != NULL) {
                BITMAP *bmp;
                int posx = x * 50;
                int posy = y * 50;
                if (mapa[x][y]->andando) {
                    switch (mapa[x][y]->direcao) {
                        case CIMA: posy += 25; break;
                        case BAIXO: posy -= 25; break;
                        case ESQUERDA: posx += 25; break;
                        case DIREITA: posx -= 25; break;
                    }
                }
                bmp = (mapa[x][y]->andando) ? (mapa[x][y]->imagem[0]) : (mapa[x][y]->imagem[1]);
                int rotacao = rotacionar_objeto(mapa[x][y]->direcao);
                rotate_sprite(buffer, bmp, posx, posy, itofix(rotacao));
            }
        }
    }
    blit(buffer, screen, 0, 0, 0, 0, MAX_X, MAX_Y);
}

/**
* Executa o jogo enquanto nao for pressionada a tecla ESC
*/
void executar() {
    while (!sair) {
        for (int x = 0; x < 16; x++) {
            for (int y = 0; y < 12; y++) {
                executar_elemento(x, y);
            }
        }
        desenhar();
        passo = (passo == 0) ? (1) : (0);
        if (key[KEY_ESC]) {
            sair = true;
        }
        clear_keybuf();
        rest(100);
        jogada++;
    }
}

/**
* Le o arquivo de mapa e constroi o ambiente com seus elementos
*/
void construir() {
    FILE *file;
    if ((file = fopen("tabuleiro.map", "r")) == NULL) {
        printf ("Arquivo nao pode ser aberto\n");
        exit(1);
    }
    BITMAP *img_parede = load_bitmap("parede.bmp", NULL);
    BITMAP *img_monstro0 = load_bitmap("monstro0.bmp", NULL);
    BITMAP *img_monstro1 = load_bitmap("monstro1.bmp", NULL);
    BITMAP *img_avatar0 = load_bitmap("avatar0.bmp", NULL);
    BITMAP *img_avatar1 = load_bitmap("avatar1.bmp", NULL);

    int x = 0;
    int y = 0;
    while (!feof(file)) {
        Elemento* e = NULL;
        bool caracterValido = true;
        int item = fgetc(file);
        switch (item) {
            case '0':
                break;
            case '1':
                mapa[x][y] = new Elemento(PAREDE, img_parede, img_parede);
                break;
            case '2':
                mapa[x][y] = new Elemento(MONSTRO, img_monstro0, img_monstro1);
                break;
            case '3':
                mapa[x][y] = new Elemento(AVATAR, img_avatar0, img_avatar1);
                break;
            default:
                caracterValido = false;
        }
        if (caracterValido) {
            x++;
            if (x >= 16) {
                x = 0;
                y++;
            }
        }
    }
    fclose(file);
}

/**
* Inicializa o allegro, variaveis e o ambiente
*/
void inicializar() {
    allegro_init();
    install_keyboard();
    install_timer();
    set_color_depth(16);
    set_gfx_mode(VMODE, MAX_X, MAX_Y, 0, 0);
    tempo = 0;
    install_int(aumenta_tempo, 1000);
    LOCK_VARIABLE(tempo);
    LOCK_FUNCTION(aumenta_tempo);

    buffer = create_bitmap(MAX_X, MAX_Y);
    chao = load_bitmap("chao.bmp", NULL);
    passo = 0;
    jogada = 0;
    sair = false;
    construir();
}

/**
* Finaliza os elementos e o allegro
*/
void finalizar() {
    for (int x = 0; x < 16; x++) {
        for (int y = 0; y < 12; y++) {
            if (mapa[x][y] != NULL)
                delete mapa[x][y];
        }
    }
    remove_int(aumenta_tempo);
    destroy_bitmap(chao);
    destroy_bitmap(buffer);
    allegro_exit();
}

/**
* Funcao principal
*/
int main () {
    inicializar();
    executar();
    finalizar();
    return 0;
}
END_OF_MAIN();
